from . import metadata, shell
from ._client import Client
from ._enums import CampLevel, Castle, CrewRole, EventReservationOption, Season, SourcePoll, TShirtSize, World
from ._exceptions import HTTPClientLookupError, InvalidGalleryError, RevoteError
from ._http import HTTPClient
from ._models import (
    Camp,
    CastleGalleries,
    Child,
    CrewMember,
    EventReservation,
    Gallery,
    PersonalReservationInfo,
    Photo,
    PlebisciteCandidate,
    Purchaser,
    Reservation,
    Resource,
    Transport
)

__all__ = (
    "Client",
    "HTTPClient",
    "CrewRole",
    "Castle",
    "CampLevel",
    "World",
    "Season",
    "EventReservationOption",
    "TShirtSize",
    "SourcePoll",
    "RevoteError",
    "InvalidGalleryError",
    "HTTPClientLookupError",
    "Resource",
    "Gallery",
    "Camp",
    "CastleGalleries",
    "Purchaser",
    "PersonalReservationInfo",
    "Reservation",
    "EventReservation",
    "CrewMember",
    "PlebisciteCandidate",
    "Photo",
    "Transport",
    "Child",
    "shell",
    "metadata"
)



"""
Pokazuje imię i zdjęcie psora, który ma najwięcej wystąpień wylosowanej litery w opisie (wiem, bardzo kreatywnie).
"""

from asyncio import run
from io import BytesIO
from random import choice
from string import ascii_lowercase

from PIL import Image  # pip install pillow

from pymondis import Client, CrewRole


async def main():
    async with Client() as client:
        crew = await client.get_crew()
        psor_list = filter(  # 'psors' nie brzmi dobrze...
            lambda crew_member: crew_member.position == CrewRole.PSOR,
            crew
        )  # Filtruje wszystkich kierowników (no, na pozycji nie-psora)
        random_letter = choice(ascii_lowercase)
        psor_list_sorted = sorted(
            psor_list,
            key=lambda psor: psor.description.lower().count(random_letter),
            reverse=True
        )  # Sortuje psorów wg. rosnącej ilości jej wystąpień
        psor = psor_list_sorted[0]  # Pierwszy psor z listy ma najwięcej wystąpień szukanej litery
        photo = await psor.photo.get()  # Pobieranie zdjęcia
        print("Litera: {}, Psor: {} {}{}, Wystąpienia: {}"
              .format(random_letter, psor.name, psor.surname,
                      "" if psor.character is None else " ({})".format(psor.character),
                      psor.description.lower().count(random_letter))
              )
        Image.open(BytesIO(photo)).show()


if __name__ == "__main__":
    run(main())



"""
Pokazuje listę masterskich obozów od najmniejszej ceny.
"""


from asyncio import run

from pymondis import CampLevel, Client


async def main():
    async with Client() as client:
        camps = await client.get_camps()  # Pobiera listę obozów
        master_camps = filter(lambda camp: camp.level is CampLevel.MASTER, camps)  # Filtruje zostawiając masterskie
        master_camps_sorted = sorted(master_camps, key=lambda camp: camp.promo or camp.price)  # Sortuje wg. ceny
        for camp in master_camps_sorted:
            print(camp)

if __name__ == "__main__":
    run(main())


from typing import Callable, NoReturn

from httpx import AsyncClient

from ._util import default_backoff
from .metadata import __title__ as project_name, __version__ as project_version


class HTTPClient(AsyncClient):
    """
    Podstawowa klasa bezpośrednio wykonująca zapytania.
    Zwraca surowe dane.
    Jest używana wewnętrznie przez klasę ``Client`` (zalecane jest korzystanie tylko z niego).
    """
    def __init__(
            self,
            timeout: float | None = None,
            backoff: Callable[[Callable], Callable] = default_backoff,
            *,
            base_url: str = "https://quatromondisapi.azurewebsites.net/api"
    ):
        """
        Initializuje instancję.

        :param timeout: czas, po którym client zostanie samoistnie rozłączony gdy, nie uzyska odpowiedzi.
        :param base_url: podstawowy url, na który będą kierowane zapytania (z wyjątkiem ``get_resource``).
        """
        super().__init__(timeout=timeout, base_url=base_url)
        self.headers = {"User-Agent": "{}/{}".format(project_name, project_version)}

        self.send = backoff(self.send)

    async def get_api_camps(self) -> list[dict[str, str | int | bool | None | list[str | dict[str, str | int]]]]:
        """
        Zwraca dane o aktualnych obozach.

        :returns: lista informacji o obozach.
        """
        response = await self.get(
            "/Camps",
            headers={"Accept": "application/json"}
        )
        return response.json()

    async def post_api_events_inauguration(self, reservation_model: dict):
        """
        Rezerwuje inaugurację.

        :param reservation_model: dane o rezerwacji.
        """
        await self.post(
            "/Events/Inauguration",
            json=reservation_model
        )

    async def get_api_images_galleries_castles(self) -> list[dict[str, str | int | bool]]:
        """
        Podaje czy dany zamek jest aktywny pod kątem galerii.

        :returns: lista ze statusami zamków.
        """
        response = await self.get(
            "/Images/Galeries/Castles",  # 'Galeries' - English 100
            headers={"Accept": "application/json"})

        return response.json()

    async def get_api_images_galleries_castle(self, castle: str) -> list[dict[str, str | int | bool]]:
        """
        Dostaje podstawowe dane na temat aktualnych galerii z danego zamku.

        :param castle: nazwa zamku, z którego pobierana jest lista galerii.
        :returns: lista reprezentująca aktualne galerie z zamku.
        """
        response = await self.get(
            "/Images/Galeries/Castle/{}".format(castle),  # Znowu 'Galeries'
            headers={"Accept": "application/json"})

        return response.json()

    async def get_api_images_galleries(self, gallery_id: int) -> list[dict[str, str]]:
        """
        Dostaje linki do zdjęć znajdujących się w galerii o danym ID.

        :param gallery_id: numer/ID galerii.
        :returns: lista linków do zdjęć w dwóch jakościach.
        """
        response = await self.get(
            "/Images/Galeries/{}".format(gallery_id),  # 'Galeries'...
            headers={"Accept": "application/json"})

        return response.json()

    async def post_api_orders_fourworldsbeginning(self, purchaser: dict):
        """
        Zamawia książkę „QUATROMONDIS – CZTERY ŚWIATY HUGONA YORCKA. OTWARCIE”.

        :param purchaser: dane o osobie zamawiającej.
        """
        await self.post(
            "/Orders/FourWorldsBeginning",
            json=purchaser
        )

    async def post_api_parentszone_survey(self, survey_hash: str, result: dict):
        """
        Prawdopodobnie nieobowiązujący już endpoint do jakiejś ankiety.

        :param survey_hash: ?
        :param result: opinia na temat obozu/obozów (?).
        """
        await self.post(
            "/ParentsZone/Survey/{}".format(survey_hash),
            json=result
        )

    async def get_api_parentszone_crew(self) -> list[dict[str, str]]:
        """
        Zwraca dane wszystkich psorów i kierowników.

        :returns: lista danych o kadrze
        """
        response = await self.get(
            "/ParentsZone/Crew",
            headers={"Accept": "application/json"}
        )

        return response.json()

    async def post_api_parentszone_apply(self):
        """
        Zgłasza cię do pracy.

        :raises ``NotImplementedError``: zawsze, bo metoda nie jest zaimplementowana -.-
        """
        raise NotImplementedError(
            "Ta metoda nie jest jeszcze zaimplementowana."
            "Zamiast niej możesz skorzystać z tradycyjnego formularza na stronie, śledząc wysyłane zapytania - "
            "może devtools w tab-ie NETWORK czy coś innego (nie znam się)."
            "Pamiętaj żeby nie wysyłać niczego gdy rzeczywiście nie chcesz zgłosić się do pracy."
            "Później otwórz nowy issue (https://github.com/Asapros/pymondis/issues (Implementacja zapytania POST)"
            "i podziel się nagranym zapytaniem (nie zapomnij za cenzurować danych osobowych)"
        )
        # Dane najprawdopodobniej są wysyłane jako form, ale nie ma tego w swagger-ze, a ja jestem borowikiem w
        # javascript-a i nie czaje, o co chodzi, dodajcie do dokumentacji pls

    async def post_api_reservations_subscribe(self, reservation_model: dict) -> list[str]:
        """
        Rezerwuje obóz.

        :param reservation_model: dane o osobie rezerwującej.
        :returns: lista kodów rezerwacji.
        """
        response = await self.post(
            "/Reservations/Subscribe",
            json=reservation_model,
            headers={"Accept": "application/json"}
        )

        return response.json()

    async def post_api_reservations_manage(self, pri: dict[str, str]) -> dict[str, str | bool]:
        """
        Dostaje dane o rezerwacji na podstawie jej kodu i nazwiska osoby rezerwującej.

        :param pri: kod i nazwisko.
        :returns: dokładniejsze dane o rezerwacji.
        """
        response = await self.post(
            "/Reservations/Manage",
            json=pri,
            headers={"Accept": "application/json"}
        )

        return response.json()

    async def patch_api_vote(self, category: str, name: str):
        """
        Głosuje na kandydata plebiscytu.

        :param category: kategoria, w której startuje kandydat.
        :param name: nazwa kandydata (najczęściej nazwisko).
        """
        await self.patch(  # A mnie dalej zastanawia, czemu tu patch jest, a nie post...
            "/Vote/{}/{}".format(category, name)
        )

    async def get_api_vote_plebiscite(self, year: int) -> list[dict[str, str | int | bool]]:
        """
        Zwraca surowe dane o kandydatach plebiscytu z danego roku (bez opisów :/).

        :param year: rok, z którego szukani są kandydaci (>= 2019).
        :returns: lista reprezentująca kandydatów.
        """
        response = await self.get(
            "/Vote/plebiscite/{}".format(year),  # Jedyny endpoint nie w PascalCase
            headers={"Accept": "application/json"}
        )

        return response.json()

    async def __aenter__(self) -> "HTTPClient":  # Type-hinting
        await super().__aenter__()
        return self

    def __enter__(self) -> NoReturn:
        raise RuntimeError("'with' nie można używać na HTTPClient'cie. Może chodziło ci o 'async with'?")

    def __exit__(self, exc_type, exc_val, exc_tb):
        return



"""
Dane o projekcie (potrzebne w setup.py i w generowaniu User-Agenta).
"""
__all__ = "__title__", "__version__", "__description__", "__author__", "__license__"
__title__ = "pymondis"
__version__ = "1.0.0.b2.post1"
__description__ = "Nieoficjalny wrapper Quatromondis API"
__author__ = "Asapros"
__license__ = "MIT"



"""
Przydatne funkcje.
"""
from asyncio import run, sleep
from datetime import datetime
from functools import wraps

from httpx import AsyncClient, ConnectError, HTTPStatusError

from ._enums import EventReservationOption
from ._exceptions import HTTPClientLookupError


def default_backoff(function):
    """
    Dekorator funkcji wykonujących zapytania.

    :param function: funkcja do wrapowania.
    :returns: wrapowana funkcja.
    :raises HTTPStatusError: nie udało się pomyślnie wykonać zapytania (kod błędu 400-499 lub 3 próby zakończone >= 500).
    :raises ConnectError: problem z połączeniem.
    """

    @wraps(function)
    async def inner_backoff(*args, **kwargs):
        tries: int = 0
        while True:
            tries += 1
            try:
                response = await function(*args, **kwargs)
                response.raise_for_status()
                return response
            except HTTPStatusError as error:
                if tries >= 3:
                    raise
                if not error.response.is_server_error:
                    if error.response.is_client_error:
                        raise
                    return error.response
            except ConnectError:
                if tries >= 3:
                    raise

            await sleep(tries + 0.5)

    return inner_backoff


def synchronize(function):
    """
    Dekorator do zamiany funkcji asynchronicznych na synchroniczne.

    :param function: funkcja do wrapowania.
    :returns: wrapowana funkcja.
    """
    @wraps(function)
    def inner_synchronize(*args, **kwargs):
        return run(function(*args, **kwargs))
    return inner_synchronize


def choose_http(*http_clients: AsyncClient | None):
    for http_client in http_clients:
        if http_client is None or http_client.is_closed:
            continue
        return http_client
    raise HTTPClientLookupError()


def datetime_from_string(value: str) -> datetime:
    """
    Zamienia stringa na datetime (%Y-%m-%dT%H:%M:%S).

    :param value: string do zamiany.
    :returns: datetime ze stringa.
    """
    return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")


def string_from_datetime(value: datetime) -> str:
    """
    Zamienia datetime na stringa (%Y-%m-%dT%H:%M:%S).

    :param value: datetime do zamiany.
    :returns: string z datetime.
    """
    return value.strftime("%Y-%m-%dT%H:%M:%S")


def datetime_converter(value: str | datetime) -> datetime:
    """
    Zamienia stringa na datetime (%Y-%m-%dT%H:%M:%S) jeśli to potrzebne.

    :param value: string do ewentualnej zamiany lub datetime.
    :returns: datetime ze stringa lub podany datetime.
    """
    return value if isinstance(value, datetime) else datetime_from_string(value)


def optional_character_converter(value: str) -> str | None:
    """
    Zamienia string-a "Nazwa postaci Quatromondis" na None.
    (ktoś stwierdził, że taki będzie placeholder dla imion magicznych psorów, którzy ich nie ustawili).

    :param value: string do ewentualnej zamiany.
    :returns: ``None`` jeśli string brzmiał "Nazwa postaci Quatromondis" lub podany string.
    """
    return None if value == "Nazwa postaci Quatromondis" else value


def optional_string_converter(value: str) -> str | None:
    """
    Zamienia pustego stringa na ``None``.
    (ktoś stwierdził, że taki będzie placeholder dla braku wycieczek na turnusie).

    :param value: string do ewentualnej zamiany.
    :returns: ``None`` jeśli string był pusty (długość 0) lub podany string.
    """
    return value if value else None


def price_from_ero(option: EventReservationOption) -> int:
    """
    Zamienia EventReservationOption na odpowiadającą cenę.

    :param option: opcja rezerwacji.
    :returns: cena rezerwacji.
    :raises ValueError: ta opcja nie ma przypisanej ceny.
    """
    match option:
        case EventReservationOption.CHILD:
            return 450
        case EventReservationOption.CHILD_AND_ONE_PARENT:
            return 900
        case EventReservationOption.CHILD_AND_TWO_PARENTS:
            return 1300

    raise ValueError("Opcja rezerwacji wydarzenia nie ma przypisanej ceny.")



"""
Pokazuje losowe zdjęcie z losowej galerii z losowego zamku (Może trochę zająć - 3 synchroniczne zapytania).
"""

from asyncio import run
from io import BytesIO
from random import choice

from PIL import Image  # pip install pillow

from pymondis import Client


async def main():
    async with Client() as client:
        for castle_galleries in await client.get_castles():
            if not castle_galleries.active:
                continue
            for gallery in await castle_galleries.get():
                if gallery.empty:
                    continue
                photos = await gallery.get_photos()
                if not photos:
                    continue
                photo_bytes = await choice(photos).large.get()
                break
            else:
                continue
            break
        else:
            print("Nie znaleziono żadnych zdjęć.")
        Image.open(BytesIO(photo_bytes)).show()

if __name__ == "__main__":
    run(main())



class RevoteError(Exception):
    """
    Wznoszone podczas próby zagłosowania na tą samą kategorię w inauguracji drugi raz.
    """

    def __init__(self, category: str):
        super().__init__("Próbowałeś zagłosować na kategorię '{}' drugi raz.".format(category))


class InvalidGalleryError(Exception):
    """
    Wznoszone podczas próby pobrania zdjęć z galerii będącej na liście niedziałających.
    """

    def __init__(self, gallery_id: int):
        super().__init__(
            "Galeria o ID {} jest usunięta/uszkodzona. "
            "Próba pobrania zdjęć prawdopodobnie zakończy się statusem 500 Internal Server Error.".format(gallery_id)
        )


class HTTPClientLookupError(Exception):
    """
    Wznoszone, gdy model nie posiada otartego HTTPClient-a do wykonania zapytania
    """

    def __init__(self):
        super().__init__(
            "Model nie znalazł żadnego otwartego HTTPClient-a. "
            "Jeśli nie podasz go w konstruktorze, musisz bezpośrednio w metodzie."
        )


class InactiveCastleError(Exception):
    """
    Wznoszone, podczas próby pobrania listy galerii z nieaktywnego zamku.
    """

    def __init__(self, name: str):
        super().__init__(
            "{} jest aktualnie nieaktywny i nie ma w nim żadnych galerii.".format(name)
        )



from unittest import TestCase, main

from pymondis.shell import get_camps, get_castles, get_crew, get_plebiscite


class TestClient(TestCase):
    def test_camps(self):
        get_camps()

    def test_plebiscite(self):
        get_plebiscite(2021)

    def test_castles(self):
        get_castles()

    def test_crew(self):
        get_crew()


if __name__ == "__main__":
    main()



from asyncio import gather
from datetime import datetime
from unittest import IsolatedAsyncioTestCase, main


class TestClient(IsolatedAsyncioTestCase):
    async def test_import(self): # TODO do innego plinku
        import pymondis  # A ja sobie włączyłem usuwanie niepotrzebnych importów przed commitem, dlatego tu "pass" było
        _ = pymondis.__all__

    async def test_gets(self):
        from pymondis import Client
        async with Client() as client:
            await gather(
                client.get_crew(),
                client.get_camps(),
                client.get_plebiscite(datetime.now().year)
            )

    async def test_galleries(self):
        from pymondis import Client
        async with Client() as client:
            for castle in await client.get_castles():
                if not castle.active:
                    continue
                for gallery in await castle.get():
                    if gallery.empty:
                        continue
                    await (await gallery.get_photos())[0].normal.get()
                    break
                else:
                    continue
                break
            else:
                raise ValueError("Brak aktywnych zamków do przetestowania")


if __name__ == "__main__":
    main()



from asyncio import gather, run

from aiofiles import open as aopen
from httpx import AsyncClient


async def get_scheme():
    content: bytes = b''
    async with AsyncClient() as client:
        async with client.stream("GET", "https://quatromondisapi.azurewebsites.net/swagger/docs/v1") as response:
            async for chunk in response.aiter_raw(1024):
                content += chunk
    return content


async def get_cache():
    content: bytes = b''
    try:
        async with aopen("checkdoc-cache.gz", "rb") as file:
            while chunk := await file.read(1024):
                content += chunk
    except FileNotFoundError:
        return
    return content


async def main():
    scheme, cache = await gather(get_scheme(), get_cache())
    if scheme == cache:
        print("Brak zmian w schemacie")
        return
    if cache is None:
        print("Initializowanie pliku...")
    elif scheme != cache:
        print("WYKRYTO ZMIANY W SCHEMACIE!")
    async with aopen("checkdoc-cache.gz", "wb") as file:
        await file.write(scheme)
        await file.flush()


if __name__ == "__main__":
    run(main())



"""
Trochę synchronicznych metod do użycia w konsoli.
"""
__all__ = "get_camps", "get_plebiscite", "get_castles", "get_crew", "apply_for_job"

from ._client import Client
from ._util import synchronize


@synchronize
async def get_camps():
    async with Client() as client:
        return await client.get_camps()


@synchronize
async def get_plebiscite(year: int):
    async with Client() as client:
        return await client.get_plebiscite(year)


@synchronize
async def get_castles():
    async with Client() as client:
        return await client.get_castles()


@synchronize
async def get_crew():
    async with Client() as client:
        return await client.get_crew()


@synchronize
async def apply_for_job():
    async with Client() as client:
        return await client.apply_for_job()



from setuptools import setup

from pymondis import metadata

with open("README.md", encoding="utf-8") as readme_file:
    README = readme_file.read()

with open("requirements.txt", "r") as requirements_file:
    REQUIREMENTS = requirements_file.read().splitlines()

setup(
    name=metadata.__title__,
    url="https://github.com/Asapros/pymondis",
    project_urls={
        "Tracker": "https://github.com/Asapros/pymondis/issues",
        "Source": "https://github.com/Asapros/pymondis"
    },
    version=metadata.__version__,
    packages=("pymondis",),
    license=metadata.__license__,
    author=metadata.__author__,
    description=metadata.__description__,
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=REQUIREMENTS,
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Natural Language :: Polish",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers"
    ],
    keywords=("quatromondis", "yorck", "API", "HTTP", "async", "hugo")
)



"""
Modele ułatwiające korzystanie z ``Client``a.
Wszystkie modele są tworzone za pomocą biblioteki ``attrs``.
Wszystkie modele są *frozen* - nie da się zmieniać w nich atrybutów (oprócz tych, które nie są (``Resource``)).
Wszystkie modele mają określone atrybuty (``__slots__``) co pozwala na szybszy dostęp i oszczędność do 30% pamięci.
"""

from datetime import datetime
from typing import AsyncIterator

from attr import Factory, attrib, attrs
from attr.converters import optional as optional_converter
from attr.validators import deep_iterable, instance_of as type_validator, optional as optional_validator

from ._enums import CampLevel, Castle, CrewRole, EventReservationOption, Season, SourcePoll, TShirtSize, World
from ._exceptions import InactiveCastleError, InvalidGalleryError, RevoteError
from ._http import HTTPClient
from ._util import (
    choose_http,
    datetime_converter,
    optional_character_converter,
    optional_string_converter,
    price_from_ero,
    string_from_datetime
)


@attrs(repr=True, slots=True, frozen=True, hash=True)
class ParentSurvey:
    """
    Jakaś opinia o obozach, już pewnie nie istnieje na stronach
        (na ``web.archive.org`` też nie ma).

    :ivar _http (param http): ``HTTPClient``, który będzie używany do wysłania ankiety.
    """
    _http = attrib(
        type=HTTPClient | None,
        default=None,
        validator=optional_validator(
            type_validator(HTTPClient)
        ),
        repr=False
    )

    def to_dict(self) -> dict:
        """
        Zamienia siebie na dicta.

        :returns: instancja klasy w formie dicta.
        :raises NotImplementedError: klasa nie jest zaimplementowana.
        """
        raise NotImplementedError(
            "Ta klasa jeszcze nie jest do końca zaimplementowana. "
            "Jeśli wiesz gdzie na stronie występuje form do wysłania na /api/ParentsZone/Survey/... "
            "możesz otworzyć nowy issue https://github.com/Asapros/pymondis/issues ('Implementacja zapytania POST') "
            "i się tym podzielić."
        )

    async def submit(self, survey_hash: str, http: HTTPClient | None = None):
        """
        Wrzuca ankietę.

        :param survey_hash: Dobre pytanie.
        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        """
        await choose_http(http, self._http).post_api_parentszone_survey(survey_hash, self.to_dict())


@attrs(repr=True, slots=True, frozen=True, hash=True)
class ReservationDetails:
    """
    Dokładniejsze dane o rezerwacji.
    """

    @classmethod
    def from_dict(cls, data: dict) -> "ReservationDetails":
        """
        Initializuje nową instancję za pomocą danych w dict'cie.

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :returns: instancja klasy.
        :raises NotImplementedError: klasa nie jest zaimplementowana.
        """
        raise NotImplementedError(
            "Ta klasa jeszcze nie jest do końca zaimplementowana. "
            "Jeśli masz zarezerwowany obóz i jego kod to możesz wysłać zapytanie przez "
            "HTTPClient.post_reservation_manage."
            "Otwórz nowy issue https://github.com/Asapros/pymondis/issues ('Implementacja zapytania POST') "
            "i podziel się wynikiem funkcji, nie zapomnij za cenzurować danych osobowych. "
            "Możesz też dołączyć do issue przypuszczenia do czego może być każde pole."
        )


@attrs(repr=True, slots=True)
class Resource:
    """
    Reprezentuje dane, najczęściej zdjęcie z serwera ``hymsresources.blob.core.windows.net``.

    :ivar url: link do danych.
    :ivar _http (param http): ``HTTPClient``, który będzie używany do pobrania resource'a.
    :ivar _cache_response (keyword cache_response): zapisana ostatnia odpowiedź serwera.
    """
    url = attrib(
        type=str,
        validator=type_validator(str)
    )
    _http = attrib(
        type=HTTPClient | None,
        default=None,
        validator=optional_validator(
            type_validator(HTTPClient)
        ),
        eq=False,
        repr=False
    )
    _cache_content = attrib(
        type=bytes | None,
        default=None,
        validator=optional_validator(
            type_validator(bytes)
        ),
        kw_only=True,
        eq=False,
        repr=False
    )
    _cache_etag = attrib(
        type=str | None,
        default=None,
        validator=optional_validator(
            type_validator(str)
        ),
        kw_only=True,
        eq=False,
        repr=False
    )
    _cache_last_modified = attrib(
        type=str | None,
        default=None,
        validator=optional_validator(
            type_validator(str)
        ),
        kw_only=True,
        eq=False,
        repr=False
    )

    async def get_stream(
            self,
            use_cache: bool = True,
            update_cache: bool = True,
            chunk_size: int | None = 1024,
            http: HTTPClient | None = None
    ) -> AsyncIterator[bytes]:
        """
        Otwiera strumień danych z linku.

        :param use_cache: użyć ostatniej zapisanej odpowiedzi, gdy dane się nie zmieniły?
        :param update_cache: zapisać odpowiedź do użycia później przez ``use_cache``?
        :param chunk_size: wielkość fragmentu iterowanych danych (w bajtach).
        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :returns: asynchroniczny iterator po fragmentach danych przesłanych przez serwer
        """
        # Lock-i działają wolniej w tym przypadku...
        headers = {}
        if use_cache:
            if self._cache_etag is not None:
                headers["If-None-Match"] = self._cache_etag
            if self._cache_last_modified is not None:
                headers["If-Modified-Since"] = self._cache_last_modified
        content: bytes = b""
        async with choose_http(http, self._http).stream(
            "GET",
            self.url,
            headers=headers
        ) as response:
            if response.status_code == 304:
                yield self._cache_content
                return
            async for chunk in response.aiter_bytes(chunk_size):
                content += chunk
                yield chunk
        if update_cache:
            self._cache_content = content
            self._cache_etag = response.headers["ETag"]
            self._cache_last_modified = response.headers["Last-Modified"]

    async def get(self, use_cache: bool = True, update_cache: bool = True, http: HTTPClient | None = None) -> bytes:
        """
        Całkowicie pobiera dane z linku.

        :param use_cache: użyć ostatniej zapisanej odpowiedzi, gdy dane się nie zmieniły?
        :param update_cache: zapisać odpowiedź do użycia później przez ``use_cache``?
        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :returns: dane po całkowitym ich pobraniu przez ``get_stream``.
        """
        content: bytes = b""
        async for chunk in self.get_stream(use_cache, update_cache, None, http):
            content += chunk
        return content


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Gallery:
    """
    Reprezentuje galerię z fotorelacji.

    :cvar BLACKLIST: tuple ID usuniętych/uszkodzonych galerii.
    :ivar gallery_id: id galerii.
    :ivar start: data utworzenia galerii.
    :ivar end: data zakończenia galerii.
    :ivar name: nazwa galerii - ``Z jeśli zima + skrót zamku + numer``.
    :ivar empty: czy galeria jest pusta?
    :ivar _http (param http): ``HTTPClient``, który będzie używany do pobrania zdjęć z galerii
    """

    @attrs(repr=True, slots=True, frozen=True, hash=True)
    class Photo:
        """
        Reprezentuje zdjęcie z fotorelacji w dwóch rozdzielczościach.

        :ivar normal: zdjęcie słabej rozdzielczości.
        :ivar large: zdjęcie.
        """
        normal = attrib(
            type=Resource,
            validator=type_validator(Resource)
        )
        large = attrib(
            type=Resource,
            validator=type_validator(Resource)
        )

        @classmethod
        def from_dict(cls, data: dict, **kwargs) -> "Photo":
            r"""
            Initializuje nową instancję za pomocą danych w dict'cie.

            :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
            :param \**kwargs: dodatkowe argumenty, przekazywane dalej do konstruktorów ``Resource``.
            :returns: instancja klasy.
            """
            return cls(
                normal=Resource(data["AlbumUrl"], **kwargs),
                large=Resource(data["EnlargedUrl"], **kwargs)
            )

    BLACKLIST: tuple[int, ...] = (
        2, 6, 7, 8, 19, 20, 21, 22, 23, 24, 42, 53, 65, 69, 71, 76, 77, 85, 86, 92, 95, 107, 113, 135, 115, 129, 133
    )

    gallery_id = attrib(
        type=int,
        validator=type_validator(int)
    )
    start = attrib(
        type=datetime | None,
        converter=optional_converter(
            datetime_converter
        ),
        validator=optional_validator(
            type_validator(datetime)
        ),
        default=None
    )
    end = attrib(
        type=datetime | None,
        converter=optional_converter(
            datetime_converter
        ),
        validator=optional_validator(
            type_validator(datetime)
        ),
        default=None
    )
    name = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        ),
        default=None
    )
    empty = attrib(
        type=bool | None,
        validator=optional_validator(
            type_validator(bool)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=optional_validator(
            type_validator(HTTPClient),
        ),
        default=None,
        repr=False
    )

    async def get_photos(self, http: HTTPClient | None = None, *, ignore_blacklist: bool = False) -> list[Photo]:
        """
        Pobiera wszystkie zdjęcia z galerii.

        :param http: HTTPClient, który będzie użyty zamiast tego podanego w konstruktorze.
        :param ignore_blacklist: zignorować wystąpienie id galerii na blakliście?
        :returns: lista zdjęć.
        :raises InvalidGalleryError: galeria o tym ID najprawdopodobniej nie działa.
            (Ten wyjątek nie wzniesie się przy ``ignore_blacklist`` ustawionym na ``True``)
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        """
        if not ignore_blacklist and self.gallery_id in self.BLACKLIST:
            raise InvalidGalleryError(self.gallery_id)
        client = choose_http(http, self._http)
        photos = await client.get_api_images_galleries(self.gallery_id)
        return [
            self.Photo.from_dict(photo, http=client)
            for photo in photos
        ]

    @classmethod
    def from_dict(cls, data: dict[str, str | int | bool], **kwargs) -> "Gallery":
        r"""
        Initializuje nową instancję za pomocą danych w dict'cie.

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :param \**kwargs: dodatkowe argumenty, przekazywane dalej do konstruktora.
        :returns: instancja klasy.
        """
        return cls(
            gallery_id=data["Id"],
            start=data["StartDate"],
            end=data["EndDate"],
            name=data["Name"],
            empty=not data["HasPhotos"],
            **kwargs
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class CastleGalleries:
    """
    Reprezentuje indywidualny zamek w fotorelacji.

    :cvar _ID_TO_NAME_MAP: dict z numerami zamków jako klucze i zamkami jako wartości.
    :ivar castle: zamek.
    :ivar castle_id: ID zamku.
    :ivar active: czy zamek posiada aktywne galerie?
    :ivar _http (param http): ``HTTPClient``, który będzie używany do pobrania galerii.
    """
    castle = attrib(
        type=Castle,
        validator=type_validator(Castle)
    )
    castle_id = attrib(
        type=int | None,
        validator=optional_validator(
            type_validator(int)
        ),
        default=None,
        eq=False
    )
    active = attrib(
        type=bool | None,
        validator=optional_validator(
            type_validator(bool)
        ),
        default=None,
        eq=False
    )
    _http = attrib(
        type=HTTPClient | None,
        default=None,
        validator=optional_validator(
            type_validator(HTTPClient)
        ),
        eq=False,
        repr=False
    )

    _ID_TO_CASTLE_MAP: dict[int, Castle] = {
        1:  Castle.BARANOW,
        2:  Castle.CZOCHA,
        3:  Castle.GNIEW,
        4:  Castle.GOLUB,
        5:  Castle.KLICZKOW,
        6:  Castle.KRASICZYN,
        7:  Castle.MOSZNA,
        8:  Castle.NIDZICA,
        9:  Castle.PLUTSK,
        10: Castle.RACOT,
        11: Castle.RYBOKARTY,
        12: Castle.TUCZNO,
        13: Castle.WITASZYCE
    }

    @classmethod
    def from_dict(cls, data: dict[str, int | bool], **kwargs) -> "CastleGalleries":
        r"""
        Initializuje nową instancję za pomocą danych w dict'cie.

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :param \**kwargs: dodatkowe argumenty, przekazywane dalej do konstruktora.
        :returns: instancja klasy.
        """
        return cls(
            cls._ID_TO_CASTLE_MAP[data["Id"]],
            data["Id"],
            data["IsActive"],
            **kwargs
        )

    async def get(self, http: HTTPClient | None = None, ignore_inactivity: bool = False) -> list[Gallery]:
        """
        Dostaje listę galerii z zamku.

        :param http: HTTPClient, który będzie użyty i podany do konstruktorów zamiast tego podanego w konstruktorze.
        :param ignore_inactivity: zignorować, że zamek jest nieaktywny?
        :returns: lista aktualnych galerii z zamku.
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        :raises InactiveCastleError: ten zamek jest nieaktywny.
            (Ten wyjątek nie wzniesie się przy ``ignore_inactivity`` ustawionym na ``True``)
        """
        if not ignore_inactivity and self.active is False:
            raise InactiveCastleError(self.castle.value)
        client = choose_http(http, self._http)
        return [
            Gallery.from_dict(gallery, http=client)
            for gallery in await client.get_api_images_galleries_castle(self.castle.value)
        ]


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Camp:
    """
    Reprezentuje obóz.
    
    :ivar camp_id: id obozu.
    :ivar code: kod obozu ``Z, jeśli w zimę + skrót zamku + numer + skrót programu``
    :ivar place: zamek, w którym odbywa się obóz.
    :ivar price: cena obozu.
    :ivar promo: przeceniona cena, jeśli jest.
    :ivar active: aktywny? (nie ma listy rezerwowej).
    :ivar places_left: ilość pozostałych miejsc,
        ale jest zepsuta, bo czasem anomalnie rośnie i potrafi wynosić 75, kiedy jest lista rezerwowa...
    :ivar program: temat turnusu.
    :ivar level: poziom.
    :ivar world: świat.
    :ivar season: pora roku.
    :ivar trip: opisy wycieczki/wycieczek, jeśli jakieś są.
    :ivar start: data rozpoczęcia.
    :ivar end: data zakończenia.
    :ivar ages: lista zakresów wiekowych (ciekawe czego?).
    :ivar transports: transporty na miejsce.
    """

    @attrs(repr=True, slots=True, frozen=True, hash=True)
    class Transport:
        """
        Reprezentuje transport na obóz.

        :ivar city: nazwa miasta.
        :ivar one_way_price: cena w jedną stronę.
        :ivar two_way_price: cena w dwie strony.
        """
        city = attrib(
            type=str,
            validator=type_validator(str)
        )
        one_way_price = attrib(
            type=int,
            validator=type_validator(int)
        )
        two_way_price = attrib(
            type=int,
            validator=type_validator(int)
        )

        @classmethod
        def from_dict(cls, data: dict[str, str | int]) -> "Transport":
            """
            Initializuje nową instancję za pomocą danych w dict'cie.

            :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
            :returns: instancja klasy.
            """
            return cls(
                city=data["City"],
                one_way_price=data["OneWayPrice"],
                two_way_price=data["TwoWayPrice"]
            )

    camp_id = attrib(
        type=int,
        validator=type_validator(int)
    )
    code = attrib(
        type=str,
        validator=type_validator(str)
    )
    place = attrib(
        type=Castle,
        converter=Castle,
        validator=type_validator(Castle)
    )
    price = attrib(
        type=int,
        validator=type_validator(int)
    )
    promo = attrib(
        type=int | None,
        validator=optional_validator(
            type_validator(int)
        )
    )
    active = attrib(
        type=bool,
        validator=type_validator(bool)
    )
    places_left = attrib(
        type=int,
        validator=type_validator(int)
    )
    program = attrib(
        type=str,
        validator=type_validator(str)
    )
    level = attrib(
        type=CampLevel,
        converter=CampLevel,
        validator=type_validator(CampLevel)
    )
    world = attrib(
        type=World,
        converter=World,
        validator=type_validator(World)
    )
    season = attrib(
        type=Season,
        converter=Season,
        validator=type_validator(Season)
    )
    trip = attrib(
        type=str | None,
        converter=optional_string_converter,
        validator=optional_validator(
            type_validator(str)
        )
    )
    start = attrib(
        type=datetime,
        converter=datetime_converter,
        validator=type_validator(datetime)
    )
    end = attrib(
        type=datetime,
        converter=datetime_converter,
        validator=type_validator(datetime)
    )
    ages = attrib(
        type=list[str],
        validator=deep_iterable(
            type_validator(str)
        )
    )
    transports = attrib(
        type=list[Transport],
        validator=deep_iterable(
            type_validator(Transport)
        )
    )

    @classmethod
    def from_dict(cls, data: dict[str, str | int | bool | None | list[str | dict[str, str | int]]]) -> "Camp":
        """
        Initializuje nową instancję za pomocą danych w dict'cie

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :returns: instancja klasy.
        """
        return cls(
            data["Id"],
            data["Code"],
            data["Place"],
            data["Price"],
            data["Promo"],
            data["IsActive"],
            data["PlacesLeft"],
            data["Program"],
            data["Level"],
            data["World"],
            data["Season"],
            data["Trip"],
            data["StartDate"],
            data["EndDate"],
            data["Ages"],
            [
                cls.Transport.from_dict(transport)
                for transport in data["Transports"]
            ]
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Purchaser:
    """
    Reprezentuje osobę kupującą.

    :ivar name: imię.
    :ivar surname: nazwisko.
    :ivar email: email.
    :ivar phone: numer telefonu.
    :ivar parcel_locker: dane o paczkomacie.
    :ivar _http (param http): ``HTTPClient``, który będzie używany do wysłania zamówienia.
    """
    name = attrib(
        type=str,
        validator=type_validator(str)
    )
    surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    email = attrib(
        type=str,
        validator=type_validator(str)
    )
    phone = attrib(
        type=str,
        validator=type_validator(str)
    )
    parcel_locker = attrib(
        type=str,
        validator=type_validator(str)
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=optional_validator(
            type_validator(HTTPClient),
        ),
        default=None,
        repr=False
    )

    def to_dict(self) -> dict[str, str]:
        """
        Zamienia siebie na dicta.

        :returns: instancja klasy w formie dicta.
        """
        return {
            "Name":         self.name,
            "Surname":      self.surname,
            "Email":        self.email,
            "Phone":        self.phone,
            "ParcelLocker": self.parcel_locker
        }

    async def order_fwb(self, http: HTTPClient | None = None):
        """
        Zamawia książkę „QUATROMONDIS – CZTERY ŚWIATY HUGONA YORCKA. OTWARCIE”.

        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        """
        await choose_http(http, self._http).post_api_orders_four_worlds_beginning(self.to_dict())


@attrs(repr=True, slots=True, frozen=True, hash=True)
class PersonalReservationInfo:
    """
    Dane, za których pomocą możesz uzyskać szczegóły rezerwacji (kod i nazwisko).

    :ivar reservation_id: kod rezerwacji.
    :ivar surname: nazwisko zarezerwowanego.
    :ivar _http (param http): ``HTTPClient``, który będzie używany do dostania szczegółów rezerwacji
    """
    reservation_id = attrib(
        type=str,
        validator=type_validator(str)
    )
    surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=type_validator(HTTPClient),
        default=None,
        repr=False
    )

    def to_dict(self) -> dict[str, str]:
        """
        Zamienia siebie na dicta.

        :returns: instancja klasy w formie dicta.
        """
        return {
            "ReservationId": self.reservation_id,
            "Surname":       self.surname
        }

    async def get_details(self, http: HTTPClient | None = None) -> ReservationDetails:
        """
        Dostaje dane o rezerwacji,

        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :returns: szczegóły rezerwacji.
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        """
        return ReservationDetails.from_dict(
            await choose_http(http, self._http).post_api_reservations_manage(self.to_dict())
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Reservation:
    """
    Reprezentuje rezerwacje obozu.

    :ivar camp_id: id obozu.
    :ivar child: "główne" dziecko.
    :ivar parent_name: imię rodzica.
    :ivar parent_surname: nazwisko rodzica.
    :ivar nip: NIP rodzica.
    :ivar email: email.
    :ivar phone: numer telefonu.
    :ivar poll: źródło wiedzy o obozach.
    :ivar siblings: lista "pobocznych" dzieci.
    :ivar promo_code: kod promocyjny.
    :ivar _http (param http): ``HTTPClient``, który będzie używany do wysłania rezerwacji.
    """

    @attrs(repr=True, slots=True, frozen=True, hash=True)
    class Child:
        """
        Reprezentuje dziecko w rezerwacji.

        :ivar name: imię.
        :ivar surname: nazwisko.
        :ivar t_shirt_size: rozmiar koszulki.
        :ivar birthdate: data urodzenia.
        """
        name = attrib(
            type=str,
            validator=type_validator(str)
        )
        surname = attrib(
            type=str,
            validator=type_validator(str)
        )
        t_shirt_size = attrib(
            type=TShirtSize,
            validator=type_validator(TShirtSize)
        )
        birthdate = attrib(
            type=datetime,
            validator=type_validator(datetime)
        )

        def to_dict(self) -> dict[str, str]:
            """
            Zamienia siebie na dicta.

            :returns: instancja klasy w formie dicta.
            """
            return {
                "Name":    self.name,
                "Surname": self.surname,
                "Tshirt":  self.t_shirt_size.value,
                "Dob":     string_from_datetime(self.birthdate)
            }

    camp_id = attrib(
        type=int,
        validator=type_validator(int)
    )
    child = attrib(
        type=Child,
        validator=type_validator(Child)
    )
    parent_name = attrib(
        type=str,
        validator=type_validator(str)
    )
    parent_surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    nip = attrib(
        type=str,
        validator=type_validator(str)
    )
    email = attrib(
        type=str,
        validator=type_validator(str)
    )
    phone = attrib(
        type=str,
        validator=type_validator(str)
    )
    poll = attrib(
        type=SourcePoll,
        validator=type_validator(SourcePoll)
    )
    siblings = attrib(
        type=list[Child],
        validator=deep_iterable(
            type_validator(Child)
        ),
        factory=list
    )
    promo_code = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=optional_validator(
            type_validator(HTTPClient)
        ),
        default=None,
        repr=False
    )

    def to_dict(self) -> dict[str, int | dict[str, dict[str, str] | list[dict[str, str]]] | dict[str, str]]:
        """
        Zamienia siebie na dicta.

        :returns: instancja klasy w formie dicta.
        """
        return {
            "SubcampId": self.camp_id,
            "Childs":    {  # English 100
                "Main":     self.child.to_dict(),
                "Siblings": [sibling.to_dict() for sibling in self.siblings]
            },
            "Parent":    {
                "Name":    self.parent_name,
                "Surname": self.parent_surname,
                "Nip":     self.nip
            },
            "Details":   {
                "Email": self.email,
                "Phone": self.phone,
                "Promo": self.promo_code,
                "Poll":  self.poll.value
            }
        }

    def to_pri(self, **kwargs) -> PersonalReservationInfo:
        r"""
        Tworzy instancję ``PersonalReservationInfo`` na podstawie siebie.

        :param \**kwargs: argumenty podawane dalej do konstruktora.
        :returns: odpowiadające ``PersonalReservationInfo``.
        """
        return PersonalReservationInfo(self.camp_id, self.parent_surname, **{"http": self._http} | kwargs)

    async def reserve_camp(self, http: HTTPClient | None = None) -> list[str]:
        """
        Rezerwuje obóz na podstawie informacji w tym obiekcie.

        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :returns: lista kodów rezerwacji.
        """
        return await choose_http(http, self._http).post_api_reservations_subscribe(self.to_dict())


@attrs(repr=True, slots=True, frozen=True, hash=True)
class EventReservation:
    """
    Reprezentuje rezerwacje wydarzenia (tylko inauguracja)

    :ivar option: opcja rezerwacji.
    :ivar name: imię dziecka.
    :ivar surname: nazwisko dziecka.
    :ivar parent_name: imię rodzica.
    :ivar parent_surname: nazwisko rodzica.
    :ivar parent_reused: czy użyć ``parent_name`` i ``parent_surname`` zamiast
        ``first_parent_name`` i ``first_parent_surname``?
    :ivar phone: numer telefonu.
    :ivar email: email.
    :ivar first_parent_name: imię pierwszego rodzica do rezerwacji.
    :ivar first_parent_surname: nazwisko pierwszego rodzica do rezerwacji.
    :ivar second_parent_name: imię drugiego rodzica do rezerwacji.
    :ivar second_parent_surname: nazwisko drugiego rodzica do rezerwacji.
    ։ivar third_parent_name։ imię trzeciego rodzica do rezerwa... Dobra, żartuję.
    :ivar price: cena rezerwacji (jak ją zmienisz to *chyba* nie zarezerwujesz sobie taniej ;P).
    :ivar _http (param http): ``HTTPClient``, który będzie używany do wysłania rezerwacji.
    """
    option = attrib(
        type=EventReservationOption,
        converter=EventReservationOption,
        validator=type_validator(EventReservationOption)
    )
    name = attrib(
        type=str,
        validator=type_validator(str)
    )
    surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    parent_name = attrib(
        type=str,
        validator=type_validator(str)
    )
    parent_surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    parent_reused = attrib(
        type=bool,
        validator=type_validator(bool)
    )
    phone = attrib(
        type=str,
        validator=type_validator(str)
    )
    email = attrib(
        type=str,
        validator=type_validator(str)
    )
    first_parent_name = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        )
    )
    first_parent_surname = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        )
    )
    second_parent_name = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        )
    )
    second_parent_surname = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        )
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=type_validator(HTTPClient),
        default=None,
        repr=False
    )
    price = attrib(
        type=int,
        validator=type_validator(int),
        default=Factory(lambda self: price_from_ero(self.option), takes_self=True),
        kw_only=True
    )

    def to_dict(self) -> dict[str, str | int | bool]:
        """
        Zamienia siebie na dicta.

        :returns: instancja klasy w formie dicta.
        """
        data = {
            "Price":          self.price,
            "Name":           self.name,
            "Surname":        self.surname,
            "ParentName":     self.parent_name,
            "ParentSurname":  self.parent_surname,
            "IsParentReused": self.parent_reused,
            "Phone":          self.phone,
            "Email":          self.email
        }
        # TODO co się tutaj wydarzyło
        if self.option in (EventReservationOption.CHILD, EventReservationOption.CHILD_AND_PARENT):
            data.update(
                {"FirstParentName": self.first_parent_name, "FirstParentSurname": self.first_parent_surname}
            )
        if self.option is EventReservationOption.CHILD_AND_TWO_PARENTS:
            data.update(
                {"SecondParentName": self.second_parent_name, "SecondParentSurname": self.second_parent_surname}
            )
        return data

    async def reserve_inauguration(self, http: HTTPClient | None):
        """
        Rezerwuje inaugurację.

        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :raises HTTPClientLookupError: nie znaleziono otwartego ``HTTPClient``a.
        """
        await choose_http(http, self._http).post_api_events_inauguration(self.to_dict())


@attrs(repr=True, slots=True, frozen=True, hash=True)
class CrewMember:
    """
    Członek kadry (nie biuro ani HY).

    :ivar name: imię.
    :ivar surname: nazwisko.
    :ivar character: imię psorskie (czasem nie wpisane).
    :ivar position: rola.
    :ivar description: opis.
    :ivar photo: zdjęcie.
    """
    name = attrib(
        type=str,
        validator=type_validator(str)
    )
    surname = attrib(
        type=str,
        validator=type_validator(str)
    )
    character = attrib(
        type=str | None,
        converter=optional_character_converter,
        validator=optional_validator(
            type_validator(str)
        )
    )
    position = attrib(
        type=CrewRole,
        converter=CrewRole,
        validator=type_validator(CrewRole)
    )
    description = attrib(
        type=str,
        validator=type_validator(str)
    )
    photo = attrib(
        type=Resource,
        validator=type_validator(Resource)
    )

    @classmethod
    def from_dict(cls, data: dict[str, str], **kwargs) -> "CrewMember":
        r"""
        Initializuje nową instancję za pomocą danych w dict'cie.

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :param \**kwargs: argumenty podane do instancji zdjęcia.
        :returns: instancja klasy.
        """
        return cls(
            name=data["Name"],
            surname=data["Surname"],
            character=data["Character"].strip(),
            position=data["Position"],
            description=data["Description"],
            photo=Resource(data["PhotoUrl"], **kwargs)
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class PlebisciteCandidate:
    """
    Kandydat plebiscytu.

    :ivar name: reprezentująca nazwa (najczęściej nazwisko).
    :ivar category: kategoria, w której startuje.
    :ivar votes: liczba głosów (``None`` jeśli są ukryte).
    :ivar plebiscite: nazwa plebiscytu.
    :ivar voted: czy już dzisiaj wydano głos na tę kategorię z tego IP.
    :ivar _http (param http): ``HTTPClient``, który będzie używany do oddania głosu.
    """
    name = attrib(
        type=str,
        validator=type_validator(str)
    )
    category = attrib(
        type=str,
        validator=type_validator(str)
    )
    votes = attrib(
        type=int | None,
        validator=optional_validator(
            type_validator(int)
        ),
        default=None
    )
    plebiscite = attrib(
        type=str | None,
        validator=optional_validator(
            type_validator(str)
        ),
        default=None
    )
    voted = attrib(
        type=bool | None,
        validator=optional_validator(
            type_validator(bool)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=optional_validator(
            type_validator(HTTPClient)
        ),
        default=None,
        repr=False
    )

    @classmethod
    def from_dict(cls, data: dict[str, str | int | bool | None], **kwargs) -> "PlebisciteCandidate":
        """
        Initializuje nową instancję za pomocą danych w dict'cie.

        :param data: dict, na podstawie którego zostanie stworzona nowa instancja.
        :returns: instancja klasy.
        """
        return cls(
            name=data["Name"],
            votes=data["Result"],
            category=data["Category"],
            plebiscite=data["Plebiscite"],
            voted=data["WasVoted"],
            **kwargs
        )

    async def vote(self, http: HTTPClient | None = None, *, ignore_revote: bool = False):
        """
        Głosuje na kandydata.

        :param http: ``HTTPClient``, który będzie użyty zamiast tego podanego w konstruktorze.
        :param ignore_revote: zignorować, że głos już został oddany w tej kategorii?
        :raises RevoteError: podjęta została próba zagłosowania drugi raz na tą samą kategorię.
            (Ten wyjątek nie wzniesie się przy ``ignore_revote`` ustawionym na ``True``).
        """
        if not ignore_revote and self.voted:
            raise RevoteError(self.category)
        await choose_http(http, self._http).patch_api_vote(self.category, self.name)


Photo = Gallery.Photo
Transport = Camp.Transport
Child = Reservation.Child



from enum import Enum


class CrewRole(Enum):
    """
    Role członka kadry.
    """
    PSOR = "Tutor"
    HEADMASTER = "HeadMaster"


class Castle(Enum):
    """
    Zamki, w których organizowane są obozy.
    """
    BARANOW = "Zamek w Baranowie Sandomierskim"
    CZOCHA = "Zamek Czocha"
    GNIEW = "Zamek Gniew"
    GOLUB = "Zamek Golub Dobrzyń"
    KLICZKOW = "Zamek Kliczków"
    KRASICZYN = "Zamek w Krasiczynie"
    MOSZNA = "Zamek Moszna"
    NIDZICA = "Zamek w Nidzicy"
    PLUTSK = "Zamek w Pułtusku"
    RACOT = "Pałac Racot"
    RYBOKARTY = "Pałac Rybokarty"
    TUCZNO = "Zamek Tuczno"
    WITASZYCE = "Pałac Witaszyce"


class CampLevel(Enum):
    """
    Poziomy obozów.

    :cvar TITAN: starcie tytanów. Tak, z jakiegoś powodu to jest poziom, a nie program/świat (VARIOUS).
    """
    NORMAL = "Normal"
    MASTER = "Master"
    TITAN = "Titan"


class World(Enum):
    """
    Światy, w których organizowane są obozy.

    :cvar ALL: wszystkie 4 światy.
    :cvar VARIOUS: tematyczne turnusy, np. "Smocza Straż", "Sekret Zamkowej Krypty", "Księżniczki i Rycerze".
    """
    WIZARDS = "Wizzards"  # English 100
    PATHFINDERS = "Pathfinders"
    RECRUITS = "Recruits"
    SANGUINS = "Sanguins"

    VARIOUS = "Various"

    ALL = "All"


class Season(Enum):
    """
    Pory roku (w czterech światach są tylko dwie :P).
    """
    SUMMER = "Summer"
    WINTER = "Winter"


class EventReservationOption(Enum):
    """
    Opcje rezerwacji inauguracji.
    """
    CHILD = "Tylko dziecko"
    CHILD_AND_ONE_PARENT = "Dziecko + Rodzic"
    CHILD_AND_TWO_PARENTS = "Dziecko + 2 Rodziców"


class TShirtSize(Enum):
    """
    Rozmiary koszulki.
    """
    XS = "XS"
    S = "S"
    M = "M"
    L = "L"
    XL = "XL"


class SourcePoll(Enum):
    """
    Źródło dowiedzenia się o Quatromondis.
    """
    INTERNET = "Internet"
    SOCIALS = "Socials"
    RADIO = "Radio"
    TV = "TV"
    FRIENDS = "Friends"
    FLYERS = "Flyers"
    PRESS = "Press"


"""
Pokazuję dane wylosowanego kandydata tegorocznego plebiscytu i na niego głosuje.
"""

from asyncio import run
from datetime import datetime
from random import choice

from pymondis import Client


async def main():
    async with Client() as client:
        current_year = datetime.now().year
        candidates = await client.get_plebiscite(current_year)  # Pobiera kandydatów tegorocznego plebiscytu
        if not candidates:
            print("Nie ma jeszcze żadnych kandydatów tego roku.")
            return
        not_voted = list(
            filter(lambda candidate: not candidate.voted, candidates)
        )  # Filtruje kandydatów kategorii, na którą już dzisiaj głosowałeś
        if not not_voted:  # Lista jest pusta
            print("Zagłosowałeś już dzisiaj we wszystkich kategoriach!")
            return
        candidate = choice(not_voted)
        print("Wylosowano: {} ({} głosów, kategoria {})".format(
            candidate.name,
            "[UKRYTE]" if candidate.votes is None else candidate.votes,
            candidate.category
        ))
        input("Naciśnij enter żeby zagłosować...")
        await candidate.vote()  # Głosuje na niego


if __name__ == "__main__":
    run(main())



from typing import NoReturn

from ._http import HTTPClient
from ._models import (
    Camp,
    CastleGalleries,
    CrewMember,
    PlebisciteCandidate
)


class Client:
    """
    Pozwala na wykonywanie asynchronicznych zapytań do Quatromondis API.

    :ivar http: ``HTTPClient`` używany do wykonywania zapytań.
    """

    def __init__(self, http: HTTPClient | None = None):
        """
        Initializuje instancję Client-a.

        :param http: HTTPClient, który będzie używany zamiast tworzenia zupełnie nowego.
        """
        self.http: HTTPClient = HTTPClient() if http is None else http

    async def get_castles(self) -> list[CastleGalleries]:
        """
        Dostaje listę zamków z fotorelacji.

        :returns: lista zamków.
        """
        castles = await self.http.get_api_images_galleries_castles()
        return [CastleGalleries.from_dict(castle, http=self.http) for castle in castles]

    async def get_camps(self) -> list[Camp]:
        """
        Dostaje listę obozów.

        :returns: lista aktualnie dostępnych na stronie obozów.
        """
        camps = await self.http.get_api_camps()
        return [Camp.from_dict(camp) for camp in camps]

    async def get_crew(self) -> list[CrewMember]:
        """
        Dostaje członków kadry z obozów.

        :returns: lista psorów i kierowników.
        """
        crew = await self.http.get_api_parentszone_crew()
        return [CrewMember.from_dict(crew_member, http=self.http) for crew_member in crew]

    async def get_plebiscite(self, year: int) -> list[PlebisciteCandidate]:  # TODO Plebiscite - model
        """
        Dostaje listę kandydatów plebiscytu.

        :param year: rok, z którego szukani są kandydaci plebiscytu.
        :returns: lista kandydatów plebiscytu z podanego roku.
        """
        candidates = await self.http.get_api_vote_plebiscite(year)
        return [PlebisciteCandidate.from_dict(candidate, http=self.http) for candidate in candidates]

    async def apply_for_job(self):
        """
        Zgłasza aplikację o pracę.
        """
        await self.http.post_api_parentszone_apply()

    async def __aenter__(self) -> "Client":
        await self.http.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.http.__aexit__(exc_type, exc_val, exc_tb)

    def __enter__(self) -> NoReturn:
        self.http.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.http.__aexit__(exc_type, exc_val, exc_tb)
